/*
 * Copyright (c) 2005-2007 Substance Kirill Grouchnikov. All Rights Reserved.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 
 *  o Redistributions of source code must retain the above copyright notice, 
 *    this list of conditions and the following disclaimer. 
 *     
 *  o Redistributions in binary form must reproduce the above copyright notice, 
 *    this list of conditions and the following disclaimer in the documentation 
 *    and/or other materials provided with the distribution. 
 *     
 *  o Neither the name of Substance Kirill Grouchnikov nor the names of 
 *    its contributors may be used to endorse or promote products derived 
 *    from this software without specific prior written permission. 
 *     
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR 
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR 
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; 
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE 
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, 
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */
package org.jvnet.substance;

import java.awt.Color;
import java.awt.Component;
import java.text.DateFormat;
import java.text.NumberFormat;

import javax.swing.*;
import javax.swing.border.Border;
import javax.swing.border.EmptyBorder;
import javax.swing.table.DefaultTableCellRenderer;
import javax.swing.table.TableCellRenderer;

import org.jvnet.substance.utils.SubstanceCoreUtilities;

/**
 * Default renderer for table cells.
 * 
 * @author Kirill Grouchnikov
 */
public class SubstanceDefaultTableCellRenderer extends DefaultTableCellRenderer {
	/**
	 * Place to store the color the JLabel should be returned to after its
	 * foreground and background colors have been set to the selection
	 * background color.
	 */
	private Color unselectedForeground;

	/**
	 * Place to store the color the JLabel should be returned to after its
	 * foreground and background colors have been set to the selection
	 * background color.
	 */
	private Color unselectedBackground;

	/**
	 * Renderer for boolean columns.
	 * 
	 * @author Kirill Grouchnikov
	 */
	public static class BooleanRenderer extends JCheckBox implements
			TableCellRenderer {
		/**
		 * Border for cells that do not have focus.
		 */
		private static final Border noFocusBorder = new EmptyBorder(1, 1, 1, 1);

		/**
		 * Creates a new renderer for boolean columns.
		 */
		public BooleanRenderer() {
			super();
			this.setHorizontalAlignment(SwingConstants.CENTER);
			this.setBorderPainted(true);
		}

		/*
		 * (non-Javadoc)
		 * 
		 * @see javax.swing.table.TableCellRenderer#getTableCellRendererComponent(javax.swing.JTable,
		 *      java.lang.Object, boolean, boolean, int, int)
		 */
		public Component getTableCellRendererComponent(JTable table,
				Object value, boolean isSelected, boolean hasFocus, int row,
				int column) {
			if (isSelected) {
				this.setForeground(table.getSelectionForeground());
				// super.setBackground(table.getSelectionBackground());
			} else {
				this.setForeground(table.getForeground());
			}
			Color backgroundToSet = table.getBackground();
			if (row % 2 != 0) {
				int r = backgroundToSet.getRed();
				int g = backgroundToSet.getGreen();
				int b = backgroundToSet.getBlue();
				double coef = SubstanceCoreUtilities
						.isThemeDark(SubstanceLookAndFeel.getTheme()) ? 0.80
						: 0.96;
				backgroundToSet = new Color((int) (coef * r), (int) (coef * g),
						(int) (coef * b));
			}
			if (isSelected) {
				this.setBackground(table.getSelectionBackground());
			}
			super.setBackground(backgroundToSet);

			this.setSelected(((value != null) && ((Boolean) value)
					.booleanValue()));

			if (hasFocus) {
				this.setBorder(UIManager
						.getBorder("Table.focusCellHighlightBorder"));
			} else {
				this.setBorder(BooleanRenderer.noFocusBorder);
			}

			this.setOpaque(!SubstanceCoreUtilities.toBleedWatermark(table));

			return this;
		}
	}

	/**
	 * Renderer for icon columns.
	 * 
	 * @author Kirill Grouchnikov
	 */
	public static class IconRenderer extends SubstanceDefaultTableCellRenderer {
		/**
		 * Creates a new renderer for icon columns.
		 */
		public IconRenderer() {
			super();
			this.setHorizontalAlignment(SwingConstants.CENTER);
		}

		@Override
		public void setValue(Object value) {
			this.setIcon((value instanceof Icon) ? (Icon) value : null);
			this.setText(null);
		}
	}

	/**
	 * Renderer for number columns.
	 * 
	 * @author Kirill Grouchnikov
	 */
	public static class NumberRenderer extends
			SubstanceDefaultTableCellRenderer {
		/**
		 * Creates a new renderer for number columns.
		 */
		public NumberRenderer() {
			super();
			this.setHorizontalAlignment(SwingConstants.RIGHT);
		}
	}

	/**
	 * Renderer for double columns.
	 * 
	 * @author Kirill Grouchnikov
	 */
	public static class DoubleRenderer extends NumberRenderer {
		/**
		 * Number formatter for this renderer.
		 */
		NumberFormat formatter;

		/**
		 * Creates a new renderer for double columns.
		 */
		public DoubleRenderer() {
			super();
		}

		@Override
		public void setValue(Object value) {
			if (this.formatter == null) {
				this.formatter = NumberFormat.getInstance();
			}
			this.setText((value == null) ? "" : this.formatter.format(value));
		}
	}

	/**
	 * Renderer for date columns.
	 * 
	 * @author Kirill Grouchnikov
	 */
	public static class DateRenderer extends SubstanceDefaultTableCellRenderer {
		/**
		 * Date formatter for this renderer.
		 */
		DateFormat formatter;

		/**
		 * Creates a new renderer for date columns.
		 */
		public DateRenderer() {
			super();
		}

		@Override
		public void setValue(Object value) {
			if (this.formatter == null) {
				this.formatter = DateFormat.getDateInstance();
			}
			this.setText((value == null) ? "" : this.formatter.format(value));
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.awt.Component#setForeground(java.awt.Color)
	 */
	@Override
	public void setForeground(Color c) {
		super.setForeground(c);
		this.unselectedForeground = c;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.awt.Component#setBackground(java.awt.Color)
	 */
	@Override
	public void setBackground(Color c) {
		super.setBackground(c);
		this.unselectedBackground = c;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see javax.swing.table.TableCellRenderer#getTableCellRendererComponent(javax.swing.JTable,
	 *      java.lang.Object, boolean, boolean, int, int)
	 */
	@Override
	public Component getTableCellRendererComponent(JTable table, Object value,
			boolean isSelected, boolean hasFocus, int row, int column) {

		if (isSelected) {
			super.setForeground(table.getSelectionForeground());
			// super.setBackground(table.getSelectionBackground());
		} else {
			super
					.setForeground((this.unselectedForeground != null) ? this.unselectedForeground
							: table.getForeground());
		}
		Color backgroundToSet = (this.unselectedBackground != null) ? this.unselectedBackground
				: table.getBackground();
		if (row % 2 != 0) {
			int r = backgroundToSet.getRed();
			int g = backgroundToSet.getGreen();
			int b = backgroundToSet.getBlue();
			double coef = SubstanceCoreUtilities
					.isThemeDark(SubstanceLookAndFeel.getTheme()) ? 0.80 : 0.96;
			backgroundToSet = new Color((int) (coef * r), (int) (coef * g),
					(int) (coef * b));
		}
		super.setBackground(backgroundToSet);

		this.setFont(table.getFont());

		if (hasFocus) {
			this.setBorder(UIManager
					.getBorder("Table.focusCellHighlightBorder"));
			if (!isSelected && table.isCellEditable(row, column)) {
				Color col;
				col = UIManager.getColor("Table.focusCellForeground");
				if (col != null) {
					super.setForeground(col);
				}
				col = UIManager.getColor("Table.focusCellBackground");
				if (col != null) {
					super.setBackground(col);
				}
			}
		} else {
			this.setBorder(DefaultTableCellRenderer.noFocusBorder);
		}

		this.setValue(value);
		this.setOpaque(!SubstanceCoreUtilities.toBleedWatermark(table));
		return this;
	}
}
