/*
 * Copyright (c) 2005-2007 Substance Kirill Grouchnikov. All Rights Reserved.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 
 *  o Redistributions of source code must retain the above copyright notice, 
 *    this list of conditions and the following disclaimer. 
 *     
 *  o Redistributions in binary form must reproduce the above copyright notice, 
 *    this list of conditions and the following disclaimer in the documentation 
 *    and/or other materials provided with the distribution. 
 *     
 *  o Neither the name of Substance Kirill Grouchnikov nor the names of 
 *    its contributors may be used to endorse or promote products derived 
 *    from this software without specific prior written permission. 
 *     
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR 
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR 
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; 
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE 
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, 
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */
package org.jvnet.substance.title;

import java.awt.*;
import java.awt.geom.GeneralPath;
import java.awt.image.BufferedImage;

import javax.swing.*;

import org.jvnet.substance.SubstanceRootPaneUI;
import org.jvnet.substance.color.ColorScheme;
import org.jvnet.substance.theme.SubstanceTheme;
import org.jvnet.substance.utils.SubstanceCoreUtilities;

/**
 * Implementation of {@link SubstanceHeaderPainter} that uses "arc" painting on
 * title panes and lighter gradient near the center of the application frame.
 * 
 * @author Kirill Grouchnikov
 */
public class ArcHeaderPainter implements SubstanceHeaderPainter {
	/**
	 * The secondary painter. It is used for painting areas that do not have the
	 * painting of title pane (such as menu background, table headers etc).
	 */
	private SubstanceTitlePainter secondaryPainter;

	/**
	 * Indicates whether this painter is used to paint the toolbar background.
	 */
	private boolean isPaintingToolbars;

	/**
	 * Indicates whether this painter will cause drop shadows to appear on the
	 * top portion of the toolbars that are <b>not</b> painted by this painter.
	 */
	private boolean isPaintingToolbarDropShadows;

	/**
	 * Creates a new arc header painter that is used to paint title panes,
	 * menubars <b>and</b> toolbars.
	 */
	public ArcHeaderPainter() {
		this(true);
	}

	/**
	 * Creates a new arc header painter
	 * 
	 * @param isPaintingToolbars
	 *            If <code>true</code>, this painter will be used to paint
	 *            toolbars in addition to painting title panes and menubars.
	 */
	public ArcHeaderPainter(boolean isPaintingToolbars) {
		super();
		this.isPaintingToolbars = isPaintingToolbars;
		this.secondaryPainter = new Glass3DTitlePainter();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.jvnet.substance.painter.SubstanceGradientPainter#getDisplayName()
	 */
	public String getDisplayName() {
		return "Arc";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.jvnet.substance.title.SubstanceTitlePainter#paintTitleBackground(java.awt.Graphics2D,
	 *      java.awt.Component, int, int, int, int,
	 *      org.jvnet.substance.theme.SubstanceTheme, boolean)
	 */
	public void paintTitleBackground(Graphics2D graphics, Component comp,
			int width, int height, int leftTitleX, int rightTitleX,
			SubstanceTheme theme, boolean hasDarkBorder) {
		// create rectangular background and later draw it on
		// result image with contour clip.
		BufferedImage rectangular = SubstanceCoreUtilities.getBlankImage(width,
				height);
		Graphics2D rgraphics = (Graphics2D) rectangular.getGraphics();
		rgraphics.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
				RenderingHints.VALUE_ANTIALIAS_ON);

		ColorScheme scheme = theme.getColorScheme();
		// Color topBorderColor = SubstanceColorUtilities.getTopBorderColor(
		// scheme, scheme, 0);
		// Color midBorderColor = SubstanceColorUtilities.getMidBorderColor(
		// scheme, scheme, 0);
		// Color bottomBorderColor =
		// SubstanceColorUtilities.getBottomBorderColor(
		// scheme, scheme, 0);
		//
		// Color darkFillColor = scheme.getUltraLightColor();
		// Color midFillColor = SubstanceColorUtilities.getMidFillColor(scheme,
		// scheme, 0, false);
		// Color lightFillColor = SubstanceColorUtilities.getBottomFillColor(
		// scheme, scheme, 0, false);

		// Fill background
		GeneralPath clipTop = new GeneralPath();
		clipTop.moveTo(0, 0);
		clipTop.lineTo(width, 0);
		clipTop.lineTo(width, height / 2);
		clipTop.quadTo(width / 2, height / 4, 0, height / 2);
		clipTop.lineTo(0, 0);

		rgraphics.setClip(clipTop);
		GradientPaint gradientTopLeft = new GradientPaint(0, 0, scheme
				.getLightColor(), width / 2, 0, scheme.getUltraLightColor());
		rgraphics.setPaint(gradientTopLeft);
		rgraphics.fillRect(0, 0, width / 2, height);
		GradientPaint gradientTopRight = new GradientPaint(width / 2, 0, scheme
				.getUltraLightColor(), width, 0, scheme.getLightColor());
		rgraphics.setPaint(gradientTopRight);
		rgraphics.fillRect(width / 2, 0, width - width / 2, height);

		// Color darker = SubstanceColorUtilities.getInterpolatedColor(scheme
		// .getMidColor(), scheme.getDarkColor(), 0.8);
		GradientPaint gradientBottomLeft = new GradientPaint(0, height / 2,
				scheme.getMidColor(), width / 2, height / 2, scheme
						.getLightColor());
		GradientPaint gradientBottomRight = new GradientPaint(width / 2,
				height / 2, scheme.getLightColor(), width, height / 2, scheme
						.getMidColor());

		GeneralPath clipBottom = new GeneralPath();
		clipBottom.moveTo(0, height);
		clipBottom.lineTo(width, height);
		clipBottom.lineTo(width, height / 2);
		clipBottom.quadTo(width / 2, height / 4, 0, height / 2);
		clipBottom.lineTo(0, height);

		rgraphics.setClip(clipBottom);
		rgraphics.setPaint(gradientBottomLeft);
		rgraphics.fillRect(0, 0, width / 2, height);
		rgraphics.setPaint(gradientBottomRight);
		rgraphics.fillRect(width / 2, 0, width - width / 2, height);

		GeneralPath mid = new GeneralPath();
		mid.moveTo(width, height / 2);
		mid.quadTo(width / 2, height / 4, 0, height / 2);
		rgraphics.setPaint(gradientBottomLeft);
		rgraphics.setClip(new Rectangle(0, 0, width / 2, height));
		rgraphics.draw(mid);
		rgraphics.setPaint(gradientBottomRight);
		rgraphics
				.setClip(new Rectangle(width / 2, 0, width - width / 2, height));
		rgraphics.draw(mid);

		graphics.drawImage(rectangular, 0, 0, null);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.jvnet.substance.title.SubstanceHeaderPainter#paintExtraBackground(java.awt.Graphics2D,
	 *      java.awt.Container, java.awt.Component, int, int,
	 *      org.jvnet.substance.theme.SubstanceTheme, boolean)
	 */
	public void paintExtraBackground(Graphics2D graphics, Container parent,
			Component comp, int width, int height, SubstanceTheme theme,
			boolean hasDarkBorder) {

		JRootPane rootPane = SwingUtilities.getRootPane(parent);
		SubstanceRootPaneUI ui = (SubstanceRootPaneUI) rootPane.getUI();
		JComponent titlePane = ui.getTitlePane();

		int dx = comp.getX();
		if (titlePane != null) {
			dx += (parent.getLocationOnScreen().x - titlePane
					.getLocationOnScreen().x);
		}
		// int dy = comp.getY();

		int pWidth = (titlePane == null) ? parent.getWidth() : titlePane
				.getWidth();
		int pHeight = parent.getHeight();

		ColorScheme scheme = theme.getColorScheme();
		// System.out.println("Using " + scheme.)
		GradientPaint gradientBottomLeft = new GradientPaint(-dx, pHeight / 2,
				scheme.getMidColor(), -dx + pWidth / 2, pHeight / 2,

				scheme.getLightColor());
		GradientPaint gradientBottomRight = new GradientPaint(-dx + pWidth / 2,
				pHeight / 2, scheme.getLightColor(), -dx + pWidth, pHeight / 2,
				scheme.getMidColor());
		// GradientPaint gradientBottomLeft = new GradientPaint(-dx, pHeight /
		// 2,
		// /* scheme.getMidColor() */Color.red, -dx + pWidth / 2, pHeight / 2,
		// Color.blue/*
		// * scheme .getLightColor()
		// */);
		// GradientPaint gradientBottomRight = new GradientPaint(-dx + pWidth /
		// 2,
		// pHeight / 2, Color.blue/* scheme.getLightColor() */, -dx
		// + pWidth, pHeight / 2, Color.red/* scheme.getMidColor() */);
		// graphics.translate(-dx, 0);
		graphics.setPaint(gradientBottomLeft);
		// graphics.setColor(Color.red);
		graphics.fillRect(-dx, 0, pWidth / 2, pHeight);
		graphics.setPaint(gradientBottomRight);
		// fix for defect 195 - fill the remaining width (don't use
		// pWidth / 2 since this will leave the last pixel on odd widths).
		graphics.fillRect(-dx + pWidth / 2, 0, pWidth - pWidth / 2, pHeight);
		// graphics.translate(dx, 0);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.jvnet.substance.title.SubstanceHeaderPainter#getSecondaryPainter()
	 */
	public SubstanceTitlePainter getSecondaryPainter() {
		return this.secondaryPainter;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.jvnet.substance.title.SubstanceHeaderPainter#isPaintingToolbars()
	 */
	public boolean isPaintingToolbars() {
		return this.isPaintingToolbars;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.jvnet.substance.title.SubstanceHeaderPainter#isPaintingToolbarDropShadows()
	 */
	public boolean isPaintingToolbarDropShadows() {
		return this.isPaintingToolbarDropShadows;
	}

	/**
	 * Sets indication whether this painter will cause drop shadows to appear on
	 * the top portion of the toolbars that are <b>not</b> painted by this
	 * painter.
	 * 
	 * @param isPaintingToolbarDropShadows
	 *            Indication whether this painter will cause drop shadows to
	 *            appear on the top portion of the toolbars that are <b>not</b>
	 *            painted by this painter.
	 */
	public void setPaintingToolbarDropShadows(
			boolean isPaintingToolbarDropShadows) {
		this.isPaintingToolbarDropShadows = isPaintingToolbarDropShadows;
	}

	/**
	 * Sets indication whether this painter is used to paint the toolbar
	 * background.
	 * 
	 * @param isPaintingToolbars
	 *            Indication whether this painter is used to paint the toolbar
	 *            background.
	 */
	public void setPaintingToolbars(boolean isPaintingToolbars) {
		this.isPaintingToolbars = isPaintingToolbars;
	}
}
