/*
 * Copyright (c) 2005-2007 Substance Kirill Grouchnikov. All Rights Reserved.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 
 *  o Redistributions of source code must retain the above copyright notice, 
 *    this list of conditions and the following disclaimer. 
 *     
 *  o Redistributions in binary form must reproduce the above copyright notice, 
 *    this list of conditions and the following disclaimer in the documentation 
 *    and/or other materials provided with the distribution. 
 *     
 *  o Neither the name of Substance Kirill Grouchnikov nor the names of 
 *    its contributors may be used to endorse or promote products derived 
 *    from this software without specific prior written permission. 
 *     
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR 
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR 
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; 
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE 
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, 
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */
package org.jvnet.substance.utils;

import javax.swing.*;

import org.jvnet.substance.color.*;

/**
 * This enum is used in order to provide uniform transition effects on mouse
 * events. The effects include different visual appearance of the corresponding
 * control when the mouse hovers over it (rollover), when it's pressed or
 * selected, disabled etc.<br>
 * Each enum value represents a single state and contains information that is
 * used by the UI delegates in order to correctly paint the corresponding
 * controls. <br>
 * This class is <b>for internal use only</b>.
 * 
 * @author Kirill Grouchnikov
 */
public enum ComponentState {
	/**
	 * Disabled active. Used for disabled buttons that have been marked as
	 * <code>default</code>.
	 */
	DISABLED_ACTIVE(ColorSchemeKind.DISABLED, 0, false, false, false),

	/**
	 * Active. Used for enabled buttons that have been marked as
	 * <code>default</code>.
	 */
	ACTIVE(ColorSchemeKind.CURRENT, 0, false, true, false),

	/**
	 * Disabled selected.
	 */
	DISABLED_SELECTED(ColorSchemeKind.DISABLED, 10, true, false, false),

	/**
	 * Disabled and not selected.
	 */
	DISABLED_UNSELECTED(ColorSchemeKind.DISABLED, 0, false, false, false),

	/**
	 * Pressed selected.
	 */
	PRESSED_SELECTED(ColorSchemeKind.CURRENT, 5, true, true, true),

	/**
	 * Pressed and not selected.
	 */
	PRESSED_UNSELECTED(ColorSchemeKind.CURRENT, 8, false, true, true),

	/**
	 * Selected.
	 */
	SELECTED(ColorSchemeKind.CURRENT, 8, true, true, false),

	/**
	 * Selected and rolled over.
	 */
	ROLLOVER_SELECTED(ColorSchemeKind.CURRENT, 4, true, true, true),

	/**
	 * Not selected and rolled over.
	 */
	ROLLOVER_UNSELECTED(ColorSchemeKind.CURRENT, 3, false, true, true),

	/**
	 * Default state.
	 */
	DEFAULT(ColorSchemeKind.REGULAR, 0, false, true, false);

	/**
	 * Enum for color scheme kind. Is used in order to decouple the actual
	 * current color scheme and the decision on whether to use it.
	 * 
	 * @author Kirill Grouchnikov
	 */
	public static enum ColorSchemeKind {
		/**
		 * Current color scheme (e.g. {@link AquaColorScheme}).
		 */
		CURRENT,

		/**
		 * Regular color scheme (usually {@link MetallicColorScheme}).
		 */
		REGULAR,

		/**
		 * Disabled color scheme (usually {@link LightGrayColorScheme}).
		 */
		DISABLED
	}

	/**
	 * The corresponding color scheme kind.
	 */
	private ColorSchemeKind colorSchemeKind;

	/**
	 * The corresponding cycle count. Should be a number between 0 and 10. This
	 * number is used to compute the foreground color of some component. The
	 * color is interpolated between two values (0 corresponds to usual color,
	 * 10 corresponds to very light version of the usual color).
	 */
	private int cycleCount;

	/**
	 * Indicates whether a component with <code>this</code> state is selected.
	 */
	private boolean isSelected;

	/**
	 * Indicates whether a component with <code>this</code> state is enabled.
	 */
	private boolean isEnabled;

	/**
	 * Indicates whether a component with <code>this</code> state is in
	 * rollover state.
	 */
	private boolean isRollover;

	/**
	 * Simple constructor.
	 * 
	 * @param kind
	 *            The corresponding color scheme kind.
	 * @param count
	 *            The corresponding cycle count.
	 * @param isSelected
	 *            Indicates whether a component with <code>this</code> state
	 *            is selected.
	 * @param isEnabled
	 *            Indicates whether a component with <code>this</code> state
	 *            is enabled.
	 * @param isRollover
	 *            Indicates whether a component with <code>this</code> state
	 *            is in rollover state.
	 */
	ComponentState(ColorSchemeKind kind, int count, boolean isSelected,
			boolean isEnabled, boolean isRollover) {
		this.colorSchemeKind = kind;
		this.cycleCount = count;
		this.isSelected = isSelected;
		this.isEnabled = isEnabled;
		this.isRollover = isRollover;
	}

	/**
	 * Returns the corresponding color scheme kind
	 * 
	 * @return Corresponding color scheme kind
	 */
	public ColorSchemeKind getColorSchemeKind() {
		return this.colorSchemeKind;
	}

	/**
	 * Returns the corresponding cycle count.
	 * 
	 * @return Corresponding cycle count.
	 */
	public int getCycleCount() {
		return this.cycleCount;
	}

	/**
	 * Returns the indication of whether a component with <code>this</code>
	 * state is enabled.
	 * 
	 * @return Indication of whether a component with <code>this</code> state
	 *         is enabled.
	 */
	public boolean isEnabled() {
		return this.isEnabled;
	}

	/**
	 * Returns the indication of whether a component with <code>this</code>
	 * state is selected.
	 * 
	 * @return Indication of whether a component with <code>this</code> state
	 *         is selected.
	 */
	public boolean isSelected() {
		return this.isSelected;
	}

	/**
	 * Returns the indication of whether a component with <code>this</code>
	 * state is in rollover state.
	 * 
	 * @return Indication of whether a component with <code>this</code> state
	 *         is rollover state.
	 */
	public boolean isRollover() {
		return this.isRollover;
	}

	/**
	 * Retrieves component state based on the button model (required parameter)
	 * and button itself (optional parameter).
	 * 
	 * @param model
	 *            Button model (required).
	 * @param button
	 *            Button (optional).
	 * @return The matching component state.
	 */
	public static ComponentState getState(ButtonModel model,
			AbstractButton button) {

		// if (button instanceof JToggleButton) {
		// System.out.println(button.getWidth() + "*" + button.getHeight()
		// + ":" + (model.isEnabled() ? "enabled:" : "")
		// + (model.isArmed() ? "armed:" : "")
		// + (model.isSelected() ? "selected:" : "")
		// + (model.isRollover() ? "rollover:" : "")
		// + (button.isRolloverEnabled() ? "rollover enabled:" : "")
		// + (model.isPressed() ? "pressed:" : ""));
		// }
		//
		if (button != null) {
			if (button instanceof JButton) {
				JButton jb = (JButton) button;
				if (jb.isDefaultButton()) {
					if (model.isEnabled()) {
						if ((!model.isPressed()) && (!model.isArmed()))
							return ACTIVE;
					} else
						return DISABLED_ACTIVE;
				}
			}
		}

		if (!model.isEnabled()) {
			if (model.isSelected())
				return DISABLED_SELECTED;
			return DISABLED_UNSELECTED;
		} else if (model.isPressed() && model.isArmed()) {
			if (model.isSelected())
				return PRESSED_SELECTED;
			return PRESSED_UNSELECTED;
		} else if (model.isSelected()) {
			if (((button == null) || (button.isRolloverEnabled()))
					&& model.isRollover())
				return ROLLOVER_SELECTED;
			return SELECTED;
		} else if (((button == null) || (button.isRolloverEnabled()))
				&& model.isRollover())
			return ROLLOVER_UNSELECTED;

		return DEFAULT;
	}

}
