/*
 * Copyright (c) 2005-2007 Substance Kirill Grouchnikov. All Rights Reserved.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 
 *  o Redistributions of source code must retain the above copyright notice, 
 *    this list of conditions and the following disclaimer. 
 *     
 *  o Redistributions in binary form must reproduce the above copyright notice, 
 *    this list of conditions and the following disclaimer in the documentation 
 *    and/or other materials provided with the distribution. 
 *     
 *  o Neither the name of Substance Kirill Grouchnikov nor the names of 
 *    its contributors may be used to endorse or promote products derived 
 *    from this software without specific prior written permission. 
 *     
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR 
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR 
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; 
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE 
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, 
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */
package test.check;

import java.awt.*;

import javax.swing.*;

import org.jvnet.lafwidget.LafWidget;
import org.jvnet.lafwidget.utils.LafConstants.AnimationKind;
import org.jvnet.substance.SubstanceLookAndFeel;
import org.jvnet.substance.button.ClassicButtonShaper;
import org.jvnet.substance.painter.*;
import org.jvnet.substance.theme.*;
import org.jvnet.substance.utils.SubstanceConstants.FocusKind;
import org.jvnet.substance.utils.SubstanceConstants.Side;

import test.Check;

import com.jgoodies.forms.builder.DefaultFormBuilder;
import com.jgoodies.forms.layout.FormLayout;

public class ButtonsPanel extends JPanel {
	public JButton defaultButton;

	private interface ConfigureCommand {
		public void configure(AbstractButton ab);
	}

	private class ChainCommand implements ConfigureCommand {
		private ConfigureCommand[] commands;

		public ChainCommand(ConfigureCommand... commands) {
			super();
			this.commands = commands;
		}

		public void configure(AbstractButton ab) {
			for (ConfigureCommand cmd : this.commands)
				cmd.configure(ab);
		}
	}

	private class DisableCommand implements ConfigureCommand {
		public void configure(AbstractButton ab) {
			ab.setEnabled(false);
		}
	}

	private class SelectCommand implements ConfigureCommand {
		public void configure(AbstractButton ab) {
			if (ab instanceof JButton)
				return;
			ab.setSelected(true);
		}
	}

	private class ThemeCommand implements ConfigureCommand {
		private Object themeObj;

		public ThemeCommand(Object themeObj) {
			this.themeObj = themeObj;
		}

		public void configure(AbstractButton ab) {
			ab.putClientProperty(SubstanceLookAndFeel.THEME_PROPERTY, themeObj);
		}
	}

	private class ShaperCommand implements ConfigureCommand {
		private Object shaperObj;

		public ShaperCommand(Object shaperObj) {
			this.shaperObj = shaperObj;
		}

		public void configure(AbstractButton ab) {
			ab.putClientProperty(SubstanceLookAndFeel.BUTTON_SHAPER_PROPERTY,
					shaperObj);
		}
	}

	private class PainterCommand implements ConfigureCommand {
		private Object painterObj;

		public PainterCommand(Object painterObj) {
			this.painterObj = painterObj;
		}

		public void configure(AbstractButton ab) {
			ab.putClientProperty(
					SubstanceLookAndFeel.GRADIENT_PAINTER_PROPERTY, painterObj);
		}
	}

	private class PaintActiveCommand implements ConfigureCommand {
		public void configure(AbstractButton ab) {
			ab.putClientProperty(SubstanceLookAndFeel.PAINT_ACTIVE_PROPERTY,
					Boolean.TRUE);
		}
	}

	private class NoFocusCommand implements ConfigureCommand {
		public void configure(AbstractButton ab) {
			ab.setFocusPainted(false);
		}
	}

	private class FocusKindCommand implements ConfigureCommand {
		private FocusKind focusKind;

		public FocusKindCommand(FocusKind focusKind) {
			this.focusKind = focusKind;
		}

		public void configure(AbstractButton ab) {
			ab.putClientProperty(SubstanceLookAndFeel.FOCUS_KIND, focusKind);
		}
	}

	private class AnimationKindCommand implements ConfigureCommand {
		private AnimationKind animationKind;

		public AnimationKindCommand(AnimationKind animationKind) {
			this.animationKind = animationKind;
		}

		public void configure(AbstractButton ab) {
			ab.putClientProperty(LafWidget.ANIMATION_KIND, animationKind);
		}
	}

	private class NoMinSizeCommand implements ConfigureCommand {
		public void configure(AbstractButton ab) {
			ab.putClientProperty(
					SubstanceLookAndFeel.BUTTON_NO_MIN_SIZE_PROPERTY,
					Boolean.TRUE);
		}
	}

	private class MarginCommand implements ConfigureCommand {
		public void configure(AbstractButton ab) {
			ab.setMargin(new Insets(5, 5, 5, 5));
		}
	}

	private class StraightSideCommand implements ConfigureCommand {
		private Object sideObj;

		public StraightSideCommand(Object sideObj) {
			this.sideObj = sideObj;
		}

		public void configure(AbstractButton ab) {
			ab.putClientProperty(SubstanceLookAndFeel.BUTTON_SIDE_PROPERTY,
					this.sideObj);
		}
	}

	private class OpenSideCommand implements ConfigureCommand {
		private Object sideObj;

		public OpenSideCommand(Object sideObj) {
			this.sideObj = sideObj;
		}

		public void configure(AbstractButton ab) {
			ab.putClientProperty(
					SubstanceLookAndFeel.BUTTON_OPEN_SIDE_PROPERTY,
					this.sideObj);
		}
	}

	private class TextCommand implements ConfigureCommand {
		private String text;

		public TextCommand(String text) {
			this.text = text;
		}

		public void configure(AbstractButton ab) {
			ab.setText(this.text);
		}
	}

	private class TooltipTextCommand implements ConfigureCommand {
		private String tooltipText;

		public TooltipTextCommand(String tooltipText) {
			this.tooltipText = tooltipText;
		}

		public void configure(AbstractButton ab) {
			ab.setToolTipText(this.tooltipText);
		}
	}

	private class PopupMenuCommand implements ConfigureCommand {
		public void configure(AbstractButton ab) {
			ab.addMouseListener(new MousePopupListener(ab));
		}
	}

	private class FontCommand implements ConfigureCommand {
		private Font font;

		public FontCommand(Font font) {
			this.font = font;
		}

		public void configure(AbstractButton ab) {
			ab.setFont(this.font);
		}
	}

	private class IconCommand implements ConfigureCommand {
		private Icon icon;

		public IconCommand(Icon icon) {
			this.icon = icon;
		}

		public void configure(AbstractButton ab) {
			if ((ab instanceof JRadioButton) || (ab instanceof JCheckBox))
				return;
			ab.setIcon(this.icon);
		}
	}

	private AbstractButton[] getRow() {
		AbstractButton[] result = new AbstractButton[4];
		result[0] = new JButton("sample");
		result[1] = new JToggleButton("sample");
		result[2] = new JCheckBox("sample");
		result[3] = new JRadioButton("sample");
		return result;
	}

	private void addRow(DefaultFormBuilder builder, String label, Icon icon,
			ConfigureCommand cmd) {
		AbstractButton[] row = this.getRow();
		if (cmd != null) {
			for (AbstractButton ab : row) {
				cmd.configure(ab);
			}
		}

		JLabel jl = new JLabel(label);
		if (icon != null)
			jl.setIcon(icon);
		builder.append(jl);
		for (AbstractButton ab : row)
			builder.append(ab);
	}

	public ButtonsPanel() {
		this.setLayout(new BorderLayout());

		FormLayout lm = new FormLayout(
				"right:pref, 10dlu, left:pref:grow(1), 4dlu,"
						+ "left:pref:grow(1), 4dlu, left:pref:grow(1), "
						+ "4dlu, left:pref:grow(1)", "");
		lm.setColumnGroups(new int[][] { { 3, 5, 7, 9 } });
		DefaultFormBuilder builder = new DefaultFormBuilder(lm);
		builder.setDefaultDialogBorder();

		builder.append("");
		JLabel bLabel = new JLabel("Buttons");
		bLabel.setIcon(Check.getIcon("JButtonColor16"));
		JLabel tbLabel = new JLabel("Toggle buttons");
		tbLabel.setIcon(Check.getIcon("JToggleButtonColor16"));
		JLabel cbLabel = new JLabel("Check boxes");
		cbLabel.setIcon(Check.getIcon("JCheckBoxColor16"));
		JLabel rbLabel = new JLabel("Radio buttons");
		rbLabel.setIcon(Check.getIcon("JRadioButtonColor16"));

		bLabel.setFont(bLabel.getFont().deriveFont(Font.BOLD));
		tbLabel.setFont(rbLabel.getFont().deriveFont(Font.BOLD));
		cbLabel.setFont(cbLabel.getFont().deriveFont(Font.BOLD));
		rbLabel.setFont(rbLabel.getFont().deriveFont(Font.BOLD));

		builder.append(bLabel, tbLabel);
		builder.append(cbLabel, rbLabel);

		builder.appendSeparator("Regular settings");

		this.addRow(builder, "Enabled", null, null);
		this.addRow(builder, "Disabled", null, new DisableCommand());
		this.addRow(builder, "Selected", null, new SelectCommand());
		this.addRow(builder, "Disabled selected", null, new ChainCommand(
				new DisableCommand(), new SelectCommand()));
		this.addRow(builder, "HTML text", null, new TextCommand(
				"<html>text <b>text</b> <font color='red'>text</font>"));
		this.addRow(builder, "With tooltip", null, new TooltipTextCommand(
				"Sample tooltip"));
		this.addRow(builder, "Popup menu", null, new PopupMenuCommand());
		this.addRow(builder, "With icon", null, new IconCommand(Check
				.getIcon("flag_sweden")));

		builder.appendSeparator("Color settings");
		this.addRow(builder, "Barby Pink", null, new ThemeCommand(
				new SubstanceBarbyPinkTheme()));
		this.addRow(builder, "Bottle Green", null, new ThemeCommand(
				"Bottle Green"));
		this.addRow(builder, "Steel Blue", null, new ThemeCommand(
				new SubstanceSteelBlueTheme()));
		this.addRow(builder, "Aqua & Sunset", null, new ThemeCommand(
				new SubstanceMixTheme(new SubstanceAquaTheme(),
						new SubstanceSunsetTheme())));
		this.addRow(builder, "Green & Pink", null, new ThemeCommand(
				new SubstanceMixTheme(new SubstanceLimeGreenTheme(),
						new SubstanceBarbyPinkTheme())));
		this.addRow(builder, "Always active", null, new PaintActiveCommand());
		this.addRow(builder, "Always green", null, new ChainCommand(
				new ThemeCommand(new SubstanceBottleGreenTheme()),
				new PaintActiveCommand()));

		builder.appendSeparator("Focus indications");
		this.addRow(builder, "No focus painted", null, new NoFocusCommand());
		this
				.addRow(builder, "None", null, new FocusKindCommand(
						FocusKind.NONE));
		this
				.addRow(builder, "Text", null, new FocusKindCommand(
						FocusKind.TEXT));
		this.addRow(builder, "All", null, new FocusKindCommand(FocusKind.ALL));
		this.addRow(builder, "All inner", null, new FocusKindCommand(
				FocusKind.ALL_INNER));
		this.addRow(builder, "All strong inner", null, new FocusKindCommand(
				FocusKind.ALL_STRONG_INNER));
		this.addRow(builder, "Underline", null, new FocusKindCommand(
				FocusKind.UNDERLINE));
		this.addRow(builder, "Strong underline", null, new FocusKindCommand(
				FocusKind.STRONG_UNDERLINE));

		builder.appendSeparator("Animation speed settings");
		this.addRow(builder, "None", null, new AnimationKindCommand(
				AnimationKind.NONE));
		this.addRow(builder, "Fast", null, new AnimationKindCommand(
				AnimationKind.FAST));
		this.addRow(builder, "Regular", null, new AnimationKindCommand(
				AnimationKind.REGULAR));
		this.addRow(builder, "Slow", null, new AnimationKindCommand(
				AnimationKind.SLOW));
		this.addRow(builder, "Debug", null, new AnimationKindCommand(
				AnimationKind.DEBUG));

		builder.appendSeparator("Size settings");
		this.addRow(builder, "No min size", null, new NoMinSizeCommand());
		this.addRow(builder, "Custom margin", null, new MarginCommand());

		builder.appendSeparator("Side settings");
		this.addRow(builder, "Straight top", null, new StraightSideCommand(
				Side.TOP.name()));
		this.addRow(builder, "Straight bottom", null, new StraightSideCommand(
				Side.BOTTOM));
		this.addRow(builder, "Straight left", null, new StraightSideCommand(
				new String[] { Side.LEFT.name() }));
		this.addRow(builder, "Straight right", null, new StraightSideCommand(
				new Side[] { Side.RIGHT }));

		this.addRow(builder, "Open top", null, new ChainCommand(
				new StraightSideCommand(Side.TOP.name()), new OpenSideCommand(
						Side.TOP.name())));
		this.addRow(builder, "Open bottom", null, new ChainCommand(
				new StraightSideCommand(Side.BOTTOM), new OpenSideCommand(
						Side.BOTTOM)));
		this.addRow(builder, "Open left", null, new ChainCommand(
				new StraightSideCommand(new String[] { Side.LEFT.name() }),
				new OpenSideCommand(new String[] { Side.LEFT.name() })));
		this.addRow(builder, "Open right", null, new ChainCommand(
				new StraightSideCommand(new Side[] { Side.RIGHT }),
				new OpenSideCommand(new Side[] { Side.RIGHT })));

		builder.appendSeparator("Unicode texts");
		this.addRow(builder, "Hebrew", null, new ChainCommand(new TextCommand(
				"\u05D0\u05D1\u05D2"), new IconCommand(Check
				.getIcon("flag_israel"))));
		this.addRow(builder, "Chinese", null, new ChainCommand(new FontCommand(
				new Font("Arial Unicode MS", Font.PLAIN, 11)), new TextCommand(
				"\u4E01\u4E02\u4E03"), new IconCommand(Check
				.getIcon("flag_china"))));
		this.addRow(builder, "Cyrillic", null, new ChainCommand(
				new TextCommand("\u0430\u0431\u0432"), new IconCommand(Check
						.getIcon("flag_russia"))));
		this.addRow(builder, "Greek", null, new ChainCommand(new TextCommand(
				"\u03B1\u03B2\u03B3"), new IconCommand(Check
				.getIcon("flag_greece"))));
		this.addRow(builder, "Latin", null, new ChainCommand(new TextCommand(
				"\u00E6\u00F0\u0127\u2248"), new IconCommand(Check
				.getIcon("flag_italy"))));

		builder.appendSeparator("Shaper settings");
		this.addRow(builder, "Classic", null, new ShaperCommand(
				new ClassicButtonShaper()));
		this.addRow(builder, "Standard", null, new ShaperCommand(
				"org.jvnet.substance.button.StandardButtonShaper"));

		builder.appendSeparator("Painter settings");
		this.addRow(builder, "Classic", null, new PainterCommand(
				new ClassicGradientPainter()));
		this.addRow(builder, "Glass", null, new PainterCommand(
				"org.jvnet.substance.painter.GlassGradientShaper"));
		this.addRow(builder, "Gradient Wave", null, new PainterCommand(
				new GradientWaveGradientPainter()));
		this.addRow(builder, "Specular", null, new PainterCommand(
				new SpecularGradientPainter()));
		this.addRow(builder, "Specular Wave", null, new PainterCommand(
				new SpecularWaveGradientPainter()));
		this.addRow(builder, "Standard", null, new PainterCommand(
				new StandardGradientPainter()));
		this.addRow(builder, "Subdued", null, new PainterCommand(
				new SubduedGradientPainter()));
		this.addRow(builder, "Wave", null, new PainterCommand(
				new WaveGradientPainter()));

		JPanel panel = builder.getPanel();
		JScrollPane jsp = new JScrollPane(panel);
		panel.setOpaque(false);
		jsp.setOpaque(false);
		jsp.getViewport().setOpaque(false);
		this.add(jsp, BorderLayout.CENTER);
	}
}
